from . import instance, group, base, prototype, layout
from .context import context
from converter import utils
from sketchformat.layer_group import *

# LAYOUT_AXIS = {
#     "NONE": None,
#     "HORIZONTAL": LayoutAxis.HORIZONTAL,
#     "VERTICAL": LayoutAxis.VERTICAL,
# }

# LAYOUT_ANCHOR = {
#     "MIN": LayoutAnchor.MIN,
#     "CENTER": LayoutAnchor.MIDDLE,
#     "MAX": LayoutAnchor.MAX,
#     "BASELINE": LayoutAnchor.MIDDLE,
#     "SPACE_EVENLY": LayoutAnchor.MIDDLE,
# }


def convert(fig_symbol):
    # A symbol is an artboard with a symbolID
    master = SymbolMaster(
        **base.base_styled(fig_symbol),
        **layout.layout_information(fig_symbol),
        **prototype.prototyping_information(fig_symbol),
        symbolID=utils.gen_object_id(fig_symbol["guid"]),
    )

    # Keep the base ID as the symbol reference, create a new one for the container
    master.do_objectID = utils.gen_object_id(fig_symbol["guid"], b"symbol_master")

    # Use better names for variants if possible
    try:
        parent = context.fig_node(fig_symbol["parent"]["guid"])
        if parent and parent.get("isStateGroup", False):
            master.name = symbol_variant_name(parent, fig_symbol)

    except Exception as e:
        print(e)

    return master


def move_to_symbols_page(fig_symbol, sketch_symbol):
    # After the entire symbol is converted, move it to the Symbols page
    context.add_symbol(sketch_symbol)

    # Figma stores its stack children in bottom up order, but Sketch uses top down
    if utils.has_auto_layout(fig_symbol):
        sketch_symbol = layout.post_process_group_layout(fig_symbol, sketch_symbol)

    # Since we put the Symbol in a different page in Sketch, leave an instance where the
    # master used to be
    return instance.master_instance(fig_symbol)


def symbol_variant_name(parent, symbol):
    property_values = [x.strip().split("=")[1] for x in symbol["name"].split(",")]
    ordered_values = [parent["name"]] + property_values
    return "/".join(ordered_values)



from operator import is_
from converter import utils
from . import base, positioning, rectangle
from sketchformat.layer_group import (
    Group,
    Rect,
    AbstractStyledLayer,
    AbstractLayerGroup,
)
from sketchformat.style import *


def convert(fig_group):
    return Group(
        **base.base_styled(fig_group),
    )


def post_process_frame(fig_group: dict, sketch_group: Group) -> Group:
    convert_frame_style(fig_group, sketch_group)
    adjust_group_resizing_constraint(fig_group, sketch_group)

    return sketch_group


def adjust_group_resizing_constraint(fig_group: dict, sketch_group: Group) -> None:
    """Adjust the resizing constraint of the group to better match the .fig doc.
    Groups in .fig don't really have a resizing constraint. Instead, the children of the group resize
    relative to the parent of the group.
    If all childs have the same constraint, we can have the same behaviour in Sketch, by setting the group
    constraints to be equal to the sublayers.
    However, if there is a mix, we cannot replicate the behaviour, so we just choose some constraint and throw
    a warning"""
    if not sketch_group.layers:
        return

    first_layer_constraint = [
        sketch_group.layers[0].horizontalPins,
        sketch_group.layers[0].verticalPins,
    ]

    constraints_are_inconsistent = False
    for layer in sketch_group.layers[1:]:
        current_constraint = [layer.horizontalPins, layer.verticalPins]
        if current_constraint != first_layer_constraint:
            constraints_are_inconsistent = True
            break

    if constraints_are_inconsistent:
        utils.log_conversion_warning("GRP002", fig_group)

    sketch_group.horizontalPins = sketch_group.layers[0].horizontalPins
    sketch_group.verticalPins = sketch_group.layers[0].verticalPins


def convert_frame_style(fig_group: dict, sketch_group: AbstractLayerGroup) -> AbstractLayerGroup:
    # Convert frame styles
    # - Fill/stroke/bgblur -> Rectangle on bottom with that style
    # - Layer blur -> Rectangle with bgblur on top
    # - Shadows -> Keep in the group
    # TODO: This is one case where we could have both background blur and layer blur
    style = sketch_group.style

    # Fill and borders go on a rectangle on the bottom
    has_fills = any([f.isEnabled for f in style.fills])
    has_borders = any([b.isEnabled for b in style.borders])
    has_inner_shadows = any([b.isEnabled and b.isInnerShadow for b in style.shadows])
    has_bgblur = (
        len(style.blurs)
        and style.blurs[0].isEnabled
        and style.blurs[0].type == BlurType.BACKGROUND
    )
    has_blur = (
        len(style.blurs) and style.blurs[0].isEnabled and style.blurs[0].type == BlurType.GAUSSIAN
    )

    if has_fills or has_borders or has_bgblur:
        bgrect = rectangle.make_background_rect(fig_group, sketch_group.frame, "Frame Background")
        bgrect.style.fills = style.fills
        bgrect.style.borders = style.borders
        if has_bgblur:
            bgrect.style.blurs.append(Blur(type=BlurType.BACKGROUND, radius=style.blurs[0].radius))
        bgrect.style.shadows = style.shadows

        sketch_group.layers.insert(0, bgrect)

        style.fills = []
        style.borders = []
        style.blurs = []
        style.shadows = [s for s in style.shadows if not s.isInnerShadow]

    # Blur goes in a rectangle with bgblur at the top
    if has_blur:
        blur = rectangle.make_background_rect(
            fig_group, sketch_group.frame, f"{sketch_group.name} blur"
        )
        blur.style.blurs.append(Blur(type=BlurType.BACKGROUND, radius=style.blurs[0].radius))

        # Foreground blur, add as a layer at the top of the group
        sketch_group.layers.append(blur)
        style.blurs = []

    # Inner shadows apply to each child (if they were not put in the background rect earlier)
    # Normal shadows are untouched
    for shadow in style.shadows:
        if shadow.isInnerShadow:
            utils.log_conversion_warning("GRP001", fig_group)
            apply_inner_shadow(sketch_group, shadow)

    style.shadows = [s for s in style.shadows if not s.isInnerShadow]

    return sketch_group


def apply_inner_shadow(layer: AbstractLayerGroup, shadow: Shadow) -> None:
    for child in layer.layers:
        if isinstance(child, AbstractLayerGroup):
            apply_inner_shadow(child, shadow)
        elif isinstance(child, AbstractStyledLayer):
            child.style.shadows.append(shadow)



from typing import TypedDict, Union
from converter import utils
from sketchformat.layer_common import PaddingSelection
from sketchformat.layer_group import (
    ClippingBehavior,
    Frame,
    FlexGroupLayout,
    FreeFormGroupLayout,
    InferredGroupLayout,
    FlexDirection,
    FlexJustify,
    FlexAlign,
)


class _LayoutInformation(TypedDict, total=False):
    groupLayout: Union[FreeFormGroupLayout, InferredGroupLayout, FlexGroupLayout]
    clippingBehavior: ClippingBehavior
    leftPadding: float
    topPadding: float
    rightPadding: float
    bottomPadding: float
    paddingSelection: PaddingSelection


def layout_information(fig_frame: dict) -> _LayoutInformation:
    layout = _LayoutInformation()

    layout["clippingBehavior"] = (
        ClippingBehavior.NONE if fig_frame.get("frameMaskDisabled") else ClippingBehavior.DEFAULT
    )

    if not utils.has_auto_layout(fig_frame):
        return layout

    layout["groupLayout"] = convert_group_layout(fig_frame)

    # Set padding values from Figma frame
    layout["topPadding"] = fig_frame.get("stackVerticalPadding", 0)
    layout["rightPadding"] = fig_frame.get("stackPaddingRight", 0)
    layout["bottomPadding"] = fig_frame.get("stackPaddingBottom", 0)
    layout["leftPadding"] = fig_frame.get("stackHorizontalPadding", 0)

    # Determine padding selection type based on symmetry
    has_asymmetric_padding = (
        layout["topPadding"] != layout["bottomPadding"]
        or layout["leftPadding"] != layout["rightPadding"]
    )

    layout["paddingSelection"] = (
        PaddingSelection.INDIVIDUAL if has_asymmetric_padding else PaddingSelection.PAIRED
    )

    return layout


def convert_group_layout(fig_frame: dict) -> FlexGroupLayout:
    # Determine stack direction
    is_vertical = fig_frame["stackMode"] == "VERTICAL"
    flex_direction = FlexDirection.VERTICAL if is_vertical else FlexDirection.HORIZONTAL

    # Get spacing between items
    all_gutters_gap = fig_frame.get("stackSpacing", 0)

    # Convert alignment properties
    primary_align = fig_frame.get("stackPrimaryAlignItems", "MIN")
    counter_align = fig_frame.get("stackCounterAlignItems", "MIN")

    justify = convert_flex_justify(primary_align)
    align = convert_flex_align(counter_align)

    return FlexGroupLayout(
        flexDirection=flex_direction,
        justifyContent=justify,
        alignItems=align,
        allGuttersGap=all_gutters_gap,
    )


def convert_flex_justify(justify: str) -> FlexJustify:
    justify_mapping = {
        "MIN": FlexJustify.START,
        "CENTER": FlexJustify.CENTER,
        "MAX": FlexJustify.END,
        # We seem to have a different interpretation of "SPACE_EVENLY"
        "SPACE_EVENLY": FlexJustify.SPACE_BETWEEN,
    }

    return justify_mapping.get(justify, FlexJustify.START)


def convert_flex_align(alignment: str) -> FlexAlign:
    align_mapping = {
        "MIN": FlexAlign.START,
        "CENTER": FlexAlign.CENTER,
        "MAX": FlexAlign.END,
    }

    return align_mapping.get(alignment, FlexAlign.NONE)


def post_process_group_layout(fig_node: dict, layer_group: Frame) -> Frame:
    # If the layout has a child which is ignoring the Stack layout, and the stack
    # has a "Last on top" z-index order, we'll remove the stack layout.
    has_child_ignoring_layout = False

    for layer in layer_group.layers:
        if (
            hasattr(layer, "flexItem")
            and layer.flexItem
            and getattr(layer.flexItem, "ignoreLayout", False)
        ):
            has_child_ignoring_layout = True
            break

    if has_child_ignoring_layout and not fig_node.get("stackReverseZIndex"):
        layer_group.groupLayout = FreeFormGroupLayout()
        utils.log_conversion_warning("STK001", fig_node)
    else:
        layer_group.layers.reverse()

    return layer_group
